unit UnitMain;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes,
  Activex, UITypes, Vcl.Graphics, IOUtils, VCL.FileCtrl, ShellAPI, Web.HTTPApp,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, htcanvasdx, Vcl.ExtCtrls, htmlcomp,
  Vcl.ComCtrls, hthints, httemplates, FireDAC.Stan.Intf, FireDAC.Stan.Option,
  FireDAC.Stan.Error, FireDAC.UI.Intf, FireDAC.Phys.Intf, FireDAC.Stan.Def,
  FireDAC.Stan.Pool, FireDAC.Stan.Async, FireDAC.Phys, FireDAC.Phys.SQLite,
  FireDAC.Phys.SQLiteDef, FireDAC.Stan.ExprFuncs, FireDAC.VCLUI.Wait,
  FireDAC.Stan.Param, FireDAC.DatS, FireDAC.DApt.Intf, FireDAC.DApt,
  FireDAC.Phys.SQLiteWrapper.Stat, Data.DB, FireDAC.Comp.DataSet,
  FireDAC.Comp.Client, Vcl.StdCtrls, Generics.Collections, SyncObjs,
  htutils, htxml, htmlpars, htsearch, DMschema, DMFireDAC, htsql, htofficemail,
  htreports, htcanvas, htimage, htfiredac, htdefscriptadapter, htmldraw, htrar,
  IdContext, IdCustomHTTPServer,
  IdBaseComponent, IdComponent, IdCustomTCPServer, IdHTTPServer, htoffice;

type

  TTranslationLang = class(TDictionary<string, string>)
  public
    procedure Load(XN: THtXMLNode);
    function Translate(const s: string; EmptyIfNotFound: boolean = true): string;
  end;

  TIndexThread = class(TThread)
  public
    Indexer: THtIndexer;
    ErrorMessage: string;
    procedure Execute; override;
    destructor Destroy; override;
  end;

  TSQLiteDocumentDatabase = class(THtDocumentDatabase)
  private
    IndexProvider: TDMProvider;
    QFile, QAddFile, QError, QUpdateFile, QU, QX: TDMQuery;
    FProviderPool: TDMProviderPool;
    FAdapterPool: THtSQLAdapterPool;
  public
    constructor Create(const AConnection: string; AProviderClass: CDMProvider; ADBType: TDMProvider.TDMProviderType);
    destructor Destroy; override;
    procedure GetFolders(out AFolders: THtDocumentFolders); override;
    function GetDocumentbyFileName(const AFileName: string; out DocID: integer; out AUpdated: TDateTime): boolean; override;
    function AddDocument(const AFileName, ALocation, AFileType, ADocumentName, ATitle: string;
      FolderID: integer; ADocument: THtStorageDoc): integer; override;
    procedure DeleteDocument(DocID: integer); override;
    procedure UpdateDocument(ADocument: THtStorageDoc); override;
    procedure SetDocumentError(DocID: integer; const AMessage: string); override;
    function GetDocumentbyId(DocID: integer; var FullName, FileType, FileName, Title: string; var FolderId: integer): boolean; override;
    property ProviderPool: TDMProviderPool read FProviderPool;
  end;

  TMainForm = class(TForm)
    PageControl1: TPageControl;
    TabSheet1: TTabSheet;
    TabSheet2: TTabSheet;
    GUIPanel: THtPanel;
    Connection: TFDConnection;
    ReadTrans: TFDTransaction;
    WriteTrans: TFDTransaction;
    Query: TFDQuery;
    FDPhysSQLiteDriverLink1: TFDPhysSQLiteDriverLink;
    GUITimer: TTimer;
    IndexTimer: TTimer;
    SearchTimer: TTimer;
    Panel1: TPanel;
    Splitter1: TSplitter;
    Panel2: TPanel;
    SearchEdit: TEdit;
    SearchResults: THtPanel;
    IdHTTPServer1: TIdHTTPServer;
    ViewPanel: THtPanel;
    procedure FormCreate(Sender: TObject);
    procedure SearchEditChange(Sender: TObject);
    procedure GUITimerTimer(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure IndexTimerTimer(Sender: TObject);
    procedure SearchTimerTimer(Sender: TObject);
    procedure PageControl1Change(Sender: TObject);
    procedure SearchResultsElementClick(Sender: TElement);
    function ViewPanelGetImage(Sender: THtDocument; const Url: string;
      Element: TElement): THtBytes;
    procedure IdHTTPServer1CommandGet(AContext: TIdContext;
      ARequestInfo: TIdHTTPRequestInfo; AResponseInfo: TIdHTTPResponseInfo);
  private
    Lang: string;
    Loc: TTranslationLang;
    Thread: TThread;
    DocBase: TSQLiteDocumentDatabase;
    procedure LoadStats;
    procedure LoadLocalization;
    procedure UpdateGUI;
    function GetTranslate(Sender: THtTemplate; const s: string;
      out Res: string): boolean;
    procedure CreateSearchEngine;
    procedure DoAddFolder(const Path: string);
    procedure IndexProgress(Sender: TObject);
    procedure ThreadTerminate(Sender: TObject);
    function LocalizeHTML(const s: string; Context: THtXMLNode = nil): string;
    function GetDocument(DocID, RoleID: integer; const SearchQuery: string;
      PagedLayout: boolean = false): hstring;
  public
    SearchEngine: THtSearchEngine;
    Root, WebRoot, MainTemplate, DocTemplate: string;
    DBStats: THtXMLNode;
    CS: TCriticalSection;
    CurrentFile, CurrentFolder: string;
    Processed: integer;
    procedure AddFolder;
    procedure DeleteFolder(FolderID: integer);
    procedure StartIndexing;
    procedure OpenWeb;
  end;

var MainForm: TMainForm;

implementation

{$R *.dfm}

procedure TMainForm.CreateSearchEngine;
begin
  DocBase := TSQLiteDocumentDatabase.Create(Connection.Params.Database + '@;'+ Connection.Params.Password, TDMFDProvider, SQLite);
  SearchEngine := THtSearchEngine.Create(Root, DocBase);
  SearchEngine.StoreText := true;
  SearchEngine.ImageURL := './docimages?doc_id=%d&image=';

  SearchEngine.FileTypes.Add('DOC', ftcDocument);
  SearchEngine.FileTypes.Add('DOCX', ftcDocument);
  SearchEngine.FileTypes.Add('RTF', ftcDocument);
  SearchEngine.FileTypes.Add('XLS', ftcSpreadsheet);
  SearchEngine.FileTypes.Add('XLSX', ftcSpreadsheet);
  SearchEngine.FileTypes.Add('XLSB', ftcSpreadsheet);
  SearchEngine.FileTypes.Add('PPT', ftcPresentation);
  SearchEngine.FileTypes.Add('PPTX', ftcPresentation);
  SearchEngine.FileTypes.Add('PDF', ftcPDF);
  SearchEngine.FileTypes.Add('EPUB', ftcBook);
  SearchEngine.FileTypes.Add('FB2', ftcBook);
  SearchEngine.FileTypes.Add('TXT', ftcText);
  SearchEngine.FileTypes.Add('MD', ftcText);
  SearchEngine.FileTypes.Add('HTML', ftcText);
  SearchEngine.FileTypes.Add('EML', ftcMail);
  SearchEngine.FileTypes.Add('MSG', ftcMail);
  SearchEngine.FileTypes.Add('TBM', ftcMail);
  SearchEngine.FileTypes.Add('OLM', ftcMail);

  SearchEngine.Diacritics := THtDiacritics.Create;
  SearchEngine.Diacritics.LoadfromFile(Root + ReplaceStr('dictionary\diacritics\diacritics.json', '\', PathDelim));

  SearchEngine.FolderFactory.RegisterFolderClass('.tbb', THtBatFolder);
  SearchEngine.FolderFactory.RegisterFolderClass('.zip', THtZipFolder);
  SearchEngine.FolderFactory.RegisterFolderClass('.rar', THtRarFolder);
  SearchEngine.FolderFactory.RegisterFolderClass('.pst', TPSTFile);
  SearchEngine.FolderFactory.RegisterFolderClass('.ost', TPSTFile);

  SearchEngine.SearchResultTemplate := SearchResults.HTML.Text;
  if FileExists(WebRoot + 'thumbtemplate.txt') then
  SearchEngine.ThumbTemplate := TFile.ReadAllText(WebRoot + 'thumbtemplate.txt');
  SearchEngine.OnTranslate := GetTranslate;
  if FileExists(WebRoot + 'thumbitem.txt') then
    SearchEngine.ThumbItemTemplate := TFile.ReadAllText(WebRoot + 'thumbitem.txt');
end;

procedure TMainForm.LoadStats;
var XN, Old: THtXMLNode;
begin
  XN := HtSQLAdapter.GetXML('select file_type, count(*) xcount, sum(size/(1024*1024)) size, sum(duplicates) duplicates from documents where file_type!='''' group by 1 order by 1', []);
  Old := DBStats;
  DBStats := XN;
  Old.Free;
end;


procedure TMainForm.PageControl1Change(Sender: TObject);
begin
  if PageControl1.ActivePageIndex = 1 then
    SearchEdit.SetFocus
end;

procedure TMainForm.FormCreate(Sender: TObject);
begin
  CS := TCriticalSection.Create;
  Root := IncludeTrailingPathDelimiter(ExtractFilePath(paramstr(0)));
  WebRoot := Root + 'web\';
  MainTemplate := GUIPanel.HTML.Text;
  LoadLocalization;
  SearchResults.CurrentFile := Root;
  Connection.Params.Database := Root + 'index.sdb';
  Connection.Params.UserName := '';
  Connection.Params.Password := '';
  Connection.Connected := true;
  HtSQLAdapter := THtFireDacAdapter.Create(Connection);
  LoadStats;
  UpdateGUI;
  THtDocumentConverter.RegisterImageConverter(THtImageConverter.Create);
  CreateSearchEngine;
  SearchResults.HTML.Clear;
  IdHTTPServer1.Active := true;
  ViewPanel.LazyImageLoading := true;
  if FileExists(WebRoot + 'doctemplate.txt') then
    DocTemplate := FiletoStr(WebRoot + 'doctemplate.txt')
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  FreeAndNil(SearchEngine);
  FreeandNil(HtSQLAdapter);
  FreeandNil(CS);
  FreeandNil(Loc);
end;

procedure TMainForm.LoadLocalization;
var Idx: integer;
    LocNode: THtXMLNode;
begin
  Idx := Languages.IndexOf(SysLocale.DefaultLCID);
  if Idx >= 0 then
    Lang := AnsiLowerCase(copy(Languages.LocaleName[Idx], 1, 2))
  else
    Lang := 'en';
  if FileExists(IncludeTrailingPathDelimiter(Root + 'lang') + Lang + '.xml') then
    LocNode := THtXMLNode.CreateFromFile(IncludeTrailingPathDelimiter(Root + 'lang') + Lang + '.xml')
  else
    LocNode := THtXMLNode.CreateFromFile(IncludeTrailingPathDelimiter(Root + 'lang') + 'en.xml');
  try
    Loc := TTranslationLang.Create;
    Loc.Load(LocNode);
  finally
    LocNode.Free
  end;
end;


procedure TMainForm.SearchEditChange(Sender: TObject);
var s: string;
begin
  s := trim(SearchEdit.Text);
  if (Length(s) > 2) or ((Length(s) > 0) and IsCJK(ord(s[StringStart]))) then
  begin
    SearchTimer.Enabled := false;
    SearchTimer.Enabled := true;
  end;
end;

procedure TMainForm.SearchResultsElementClick(Sender: TElement);
begin
  if Sender['doc_id'] <> '' then
  begin
    Sender.Document.JQuery('.doc').RemoveClass('selected');
    Sender.AddClass('selected');
    SearchResults.Repaint;
    ViewPanel.HTML.Text := '<body>' + SearchEngine.GetDocument(Sender.AttrI['doc_id'], 0, SearchEdit.Text, true) + '</body>';
  end;
end;

procedure TMainForm.SearchTimerTimer(Sender: TObject);
var s, types: string;
begin
  SearchTimer.Enabled := false;
  s := trim(SearchEdit.Text);
  if s = '' then
    exit;
  types := '''DOC'',''DOCX'',''RTF'',''HTML'',''HTM'',''MD'',''TXT'',''EPUB'',''PDF'',''XLS'',''XLSX'',''XLSB'',''PPT'',''PPTX'',''OLM'',''TBM''';
  Cursor := crAppStart;
  try
    s := SearchEngine.GetSearchResult(s, types, '', true, 0);
    SearchResults.HTML.Text := s;
  finally
    Cursor := crDefault;
  end;
end;

procedure TMainForm.UpdateGUI;
var R: THtReportDocument;
begin
  R := THtReportDocument.Create;
  try
    R.Parse(MainTemplate);
    R.RegisterVariable('STATS', DBStats.NodebyName('ROWDATA'));
    R.ReportContext['FOLDER'] := CurrentFolder;
    R.ReportContext['FILE'] := CurrentFile;
    R.ReportContext['DOCUMENTS'] := Inttostr(Processed);
    Query.Open('select count(*) from folders');
    if Query.Fields[0].AsInteger > 0 then
      R.ReportContext['HASFOLDERS'] := 'true'
    else
      R.ReportContext['HASFOLDERS'] := 'false';
    if DBStats.NodebyName('ROWDATA').Count > 0 then
       R.ReportContext['HASSTATS'] := 'true'
     else
       R.ReportContext['HASSTATS'] := 'false';
    if Assigned(Thread) then
    begin
      R.ReportContext['STARTED'] := 'true';
      if Assigned(TIndexThread(Thread).Indexer) and TIndexThread(Thread).Indexer.Terminated then
        R.ReportContext['STOPPING'] := 'true'
      else
        R.ReportContext['STOPPING'] := 'false';
    end else
      R.ReportContext['STARTED'] := 'false';
    R.OnGetTemplateVar := GetTranslate;
    GUIPanel.HTML.Text := R.Render;
  finally
    R.Free
  end;
end;

function TMainForm.ViewPanelGetImage(Sender: THtDocument; const Url: string;
  Element: TElement): THtBytes;
var s, sdoc: string;
begin
  Result := nil;
  if SearchEngine = nil then
    exit;
  if pos('docimages', Url) > 0 then
  begin
    s := copy(Url, pos('?', Url), MaxInt);
    delete(s, 1, pos('=', s));
    sdoc := copy(s, 1, pos('&', s)-1);
    delete(s, 1, pos('=', s));
    Result :=  THtBytes(SearchEngine.GetDocumentImage(iStrtoInt(sdoc), 0, s))
  end else
  if FileExists(Root + Url) then
    Result := Sender.ImageFiletoBytes(Root + Url)
  else
    Result := ViewPanel.DefaultGetImage(Sender, Url, Element)
end;

function TMainForm.GetTranslate(Sender: THtTemplate; const s: string; out Res: string): boolean;
begin
  Result := Loc.ContainsKey(s);
  if Result then
    Res := Loc.Translate(s);
end;

procedure TMainForm.GUITimerTimer(Sender: TObject);
begin
  LoadStats;
  UpdateGUI;
end;

procedure TMainForm.DoAddFolder(const Path: string);
var P: TDMProvider;
    Q: TDMQuery;
begin
  P := DocBase.ProviderPool.GetProvider;
  try
    Q := P.CreateQuery(true);
    try
      Q.SelectWP([path], 'select * from folders where path=:1');
      if not Q.Dataset.IsEmpty then
        exit;
      Q.SQL := 'insert into folders(path) values (:1)';
      Q.ExecWP([path]);
      Q.Commit;
    finally
      Q.Free
    end;
  finally
    DocBase.ProviderPool.ReleaseProvider(P);
  end;
end;

procedure TMainForm.AddFolder;
var path: string;
begin
  if Vcl.FileCtrl.SelectDirectory(Loc.Translate('Add folder'), '', path, [sdNewUI, sdShowShares, sdShowEdit]) then
  begin
    DoAddFolder(path);
    UpdateGUI;
  end;
end;

procedure TMainForm.DeleteFolder(FolderID: integer);
begin
  if Assigned(Thread) then
    exit;
  Query.Open('select path from folders where folder_id=:1', [FolderID]);
  if MessageDlg(Format('Delete folder %s from list?', [Query.Fields[0].AsWideString]), TMsgDlgType.mtConfirmation,
    [TMsgDlgBtn.mbYes, TMsgDlgBtn.mbNo], 0) = mrYes then
  begin
    Cursor := crAppStart;
    try
      Query.Close;
      Query.SQL.Text := 'delete from folders where folder_id=:1';
      Query.ExecSQL;
      SearchEngine.ClearIndex;
    finally
      Cursor := crDefault;
    end;
    UpdateGUI;
  end;
end;

procedure TMainForm.StartIndexing;
var IX: THtIndexer;
    P : TDMFDProvider;
begin
  if Assigned(Thread) then
    exit;
  GUIPanel.Styles.Add('#info {display: block}');
  P := TDMFDProvider.Create(Connection.Params.Database + '@;', SQLite);
  IX := THtIndexer.Create(P, SearchEngine);
  IX.ThumbCachePath := Root + 'thumbcache' + PathDelim;
  IX.LogPath := Root;
  IX.LogFiles := true;
  IX.IndexedTypes := ' .RTF .DOC .DOCX .PPT .PPTX .XLS .XLSX .XLSB .EPUB .PDF .MSG .HTM .EML .FB2 .TBM .OLM ';
  IX.OnProgress := IndexProgress;
  Thread := TIndexThread.Create(true);
  TIndexThread(Thread).Indexer := IX;
  Thread.OnTerminate := ThreadTerminate;
  Thread.FreeOnTerminate := true;
  Thread.Resume;
  UpdateGUI;
  IndexTimer.Enabled := true;
end;

procedure TMainForm.ThreadTerminate(Sender: TObject);
begin
  if TIndexThread(Sender).ErrorMessage <> '' then
    ShowMessage(TIndexThread(Sender).ErrorMessage);
  Thread := nil;
  IndexTimer.Enabled := false;
  try
    LoadStats;
    UpdateGUI;
  except
    {}
  end;
end;

function TMainForm.LocalizeHTML(const s: string; Context: THtXMLNode = nil): string;
var T: THtTemplate;
begin
  T := THtTemplate.Create(s);
  try
    T.OnGetVariable := GetTranslate;
    if Assigned(Context) then
      T.SetVariable('.', Context);
    Result := T.Render;
  finally
    T.Free;
  end;
end;

function TMainForm.GetDocument(DocID, RoleID: integer; const SearchQuery: string; PagedLayout: boolean = false): hstring;
var T: THtTemplate;
    XN: THTXMLNode;
begin
  Result := SearchEngine.GetDocument(DocID, RoleID, SearchQuery, PagedLayout);
  if DocTemplate <> '' then
  begin
    T := THtTemplate.Create(DocTemplate);
    XN := THtXMLNode.Create('<t/>');
    try
      XN['DOC_ID'] := inttostr(DocID);
      T.SetVariable('.', XN);
      XN['LIGHT'] := 'false';
      Result := T.Render + Result;
    finally
      T.Free;
      XN.Free
    end;
  end;
end;

procedure TMainForm.IdHTTPServer1CommandGet(AContext: TIdContext;
  ARequestInfo: TIdHTTPRequestInfo; AResponseInfo: TIdHTTPResponseInfo);
var
  s, Report: string;
  HF: TStringList;
  ST: TBytesStream;
  XN: THtXMLNode;
  n: integer;
  AUpdated: TDateTime;
begin
  try
  Report := copy(string(ARequestInfo.Document), 2, MaxInt);
  if SameText(Report, 'favicon.ico') then
  begin
    AResponseInfo.ContentStream := TFileStream.Create(Webroot + Report, fmOpenRead);
    AResponseInfo.ContentType := 'image/svg+xml';
    AResponseInfo.CacheControl := 'max-age=600000';
    exit
  end;
  HF := TStringList.Create;
  try
    s := ARequestInfo.UnparsedParams;
    ExtractHeaderFields(['&'], [], s, HF, true);
    if report = '' then
    begin
      s := 'web.html';
      XN := THtXMLNode.Create('<c/>');
      try
        if ARequestInfo.Session.Content.Values['theme'] = 'light' then
          XN['LIGHT'] := 'true'
         else
           XN['LIGHT'] := 'false';
         XN['THEME'] := ARequestInfo.Session.Content.Values['theme'];
        AResponseInfo.ContentText := LocalizeHtml(FileToStr(WebRoot + s), XN);
      finally
        XN.Free
      end;
      AResponseInfo.ContentType := 'text/html;charset=utf-8';
    end else
    if AnsiStartsWith(report, 'icons/') then
    begin
      if Report  <> 'icons/.svg' then
        AResponseInfo.ContentText := FileToStr(Root + report);
      AResponseInfo.ContentType := 'image/svg+xml';
      AResponseInfo.CacheControl := 'max-age=31536000';
    end else
    if report = 'getsearch' then
    begin
      s := SearchEngine.GetSearchResult(HF.Values['text'], HF.Values['types'], HF.Values['sort'], HF.Values['allwords']='true', 0, true, HF.Values['dates']);
      AResponseInfo.ContentText := s;
      AResponseInfo.ContentType := 'text/html;charset=utf-8';
    end else
    if report = 'getthumbnails' then
    begin
      AResponseInfo.ContentText := SearchEngine.GetThumbnails(HF.Values['text'], HF.Values['types'], HF.Values['sort'], HF.Values['allwords']='true', 0);
      AResponseInfo.ContentType := 'text/html;charset=utf-8';
    end else
    if report = 'gettoc' then
    begin
      s := SearchEngine.GetTOC(iStrtoInt(HF.Values['doc_id']), 0);
      AResponseInfo.ContentText := s;
      AResponseInfo.ContentType := 'text/html;charset=utf-8';
    end else
    if report = 'getdocsearch' then
    begin
      s := SearchEngine.GetSnippets(iStrtoInt(HF.Values['doc_id']), 0, HF.Values['text']);
      AResponseInfo.ContentText := s;
      AResponseInfo.ContentType := 'text/html;charset=utf-8';
    end else
    if report = 'doc' then
    begin
      s := HF.Values['doc_id'];
      if (s = '') and (HF.Values['name'] <> '') then
      begin
        if SearchEngine.DocumentDatabase.GetDocumentbyFileName(HF.Values['name'], n, AUpdated) then
          s := Inttostr(n);
      end;
      s := '<meta charset="utf-8">' + GetDocument(iStrtoInt(s), 0, HF.Values['text'], true);
      AResponseInfo.ContentText := s;
      AResponseInfo.ContentType := 'text/html;charset=utf-8';
    end else
    if report = 'getpdf' then
    begin
      ST := TBytesStream.Create(SearchEngine.GetPDF(iStrtoInt(HF.Values['doc_id']), 0));
      AResponseInfo.ContentType := 'application/pdf';
      AResponseInfo.ContentStream := ST;
    end else
    if report = 'autocomplete' then
    begin
      s := SearchEngine.GetAutocomplete(HF.Values['term']);
      AResponseInfo.ContentText := s;
      AResponseInfo.ContentType := 'application/json;charset=utf-8';
    end else
    if report = 'docimages' then
    begin
      ST := TBytesStream.Create(SearchEngine.GetDocumentImage(iStrtoInt(HF.Values['doc_id']), 0, HF.Values['image']));
      AResponseInfo.ContentType := 'image/bitmap';
      AResponseInfo.ContentStream := ST;
      AResponseInfo.CacheControl := 'max-age=600000';
    end else
    if report = 'thumbs' then
    begin
      s := Root + 'thumbcache' + PathDelim + HF.Values['doc_id'] + '.png';
      if FileExists(s) then
        AResponseInfo.ContentStream := TFileStream.Create(s, fmOpenRead)
      else begin
        CoInitialize(nil);
        AResponseInfo.ContentStream := TBytesStream.Create(SearchEngine.GetDocumentThumbnail(iStrtoInt(HF.Values['doc_id']), 793, 0));
        ForceDirectories(Root + 'thumbcache' + PathDelim);
        TBytesStream(AResponseInfo.ContentStream).SaveToFile(s);
        TBytesStream(AResponseInfo.ContentStream).Position := 0;
      end;
      AResponseInfo.ContentType := 'image/bitmap';
      AResponseInfo.CacheControl := 'max-age=600000';
    end else
    if FileExists(WebRoot + Report) then
    begin
      AResponseInfo.CacheControl := 'max-age=86400';
      if AnsiEndsWith(report, '.ico') then
      begin
        AResponseInfo.ContentText := FileToStr(WebRoot + report);
        AResponseInfo.ContentType := 'image/icon';
      end else
      if AnsiEndsWith(report, '.png') or AnsiEndsWith(report, '.gif') then
      begin
        AResponseInfo.ServeFile( AContext,  WebRoot + report)
      end else
      begin
        AResponseInfo.ContentText := FileToStr(WebRoot + report);
        if AnsiEndsWith(report, '.css') then
          AResponseInfo.ContentType := 'text/css'
        else
        if AnsiEndsWith(report, '.js') then
          AResponseInfo.ContentType := 'text/javascript'
        else
      end;
    end else
    if report = 'getdatemap' then
    begin
      s := SearchEngine.GetDateMap(HF.Values['text'], HF.Values['allwords']='true', HF.Values['types']);
      AResponseInfo.ContentText := s;
      AResponseInfo.ContentType := 'text/html;charset=utf-8';
    end else
    if report = 'gettypemap' then
    begin
      s := SearchEngine.GetTypeMap(HF.Values['text'], HF.Values['allwords']='true', HF.Values['types']);
      AResponseInfo.ContentText := s;
      AResponseInfo.ContentType := 'text/json;charset=utf-8';
    end else
  finally
    HF.Free
  end;
  except on E: Exception do
    AResponseInfo.ContentText := '<h3>' + E.Message + '</h3>';
  end;
end;

procedure TMainForm.IndexProgress(Sender: TObject);
begin
  CS.Enter;
  try
    CurrentFile := ExtractFileName(THtIndexer(Sender).CurrentFileName);
    CurrentFolder := ExtractFilePath(THtIndexer(Sender).CurrentFileName);
    Processed := THtIndexer(Sender).Processed;
  finally
    CS.Leave
  end;
end;


procedure TMainForm.IndexTimerTimer(Sender: TObject);
begin
 GUIPanel.Doc.GetElementbyId('folder').InnerHTML := htmlpars.HtmlEncode(CurrentFolder);
 GUIPanel.Doc.GetElementbyId('file').InnerHTML := htmlpars.HtmlEncode(CurrentFile);
 GUIPanel.Doc.GetElementbyId('documents').InnerHTML := Inttostr(Processed);
 GUIPanel.Repaint;
end;

procedure OpenLinkInBrowser(const Link: string);
begin
  ShellExecute(0, 'open', pchar(Link), nil,  nil, SW_SHOWNORMAL);
end;


procedure TMainForm.OpenWeb;
begin
  OpenLinkInBrowser('http://localhost:50080')
end;


procedure TTranslationLang.Load(XN: THtXMLNode);
var i: integer;
begin
  for i := 0 to XN.Count - 1 do
    if XN.Nodes[i]['name'] <> '' then
      Add(XN.Nodes[i]['name'], XN.Nodes[i]['value']);
end;

function TTranslationLang.Translate(const s: string;
  EmptyIfNotFound: boolean): string;
begin
  if not TryGetValue(s, Result) then
  begin
    if EmptyIfNotFound then
      Result := ''
    else
      Result := s
  end;
end;

destructor TIndexThread.Destroy;
begin
  FreeandNil(Indexer);
  inherited;
end;

procedure TIndexThread.Execute;
begin
  try
    Indexer.Run;
  except on E: Exception do
    ErrorMessage := E.Message;
  end;
end;

{ TSQLiteDocumentDatabase }

function TSQLiteDocumentDatabase.AddDocument(const AFileName, ALocation,
  AFileType, ADocumentName, ATitle: string; FolderID: integer;
  ADocument: THtStorageDoc): integer;
begin
  QAddFile.SelectWP([AFileName, ALocation, AFileType, ADocument.Created, ADocument.Size, ADocumentName, ATitle, FolderID]);
  Result := QAddFile.AsInteger(0);
end;

constructor TSQLiteDocumentDatabase.Create(const AConnection: string; AProviderClass: CDMProvider; ADBType: TDMProvider.TDMProviderType);
begin
  FProviderPool := TDMProviderPool.Create(AProviderClass, AConnection, ADBType, 'JournalMode=WAL');
  FProviderPool.ReadTransactionMode := dtmDirtyRead;
  FAdapterPool := THtSQLAdapterPool.Create(AConnection);
  IndexProvider := FProviderPool.GetProvider;
  QFile := IndexProvider.CreateQuery;
  QFile.SQL := 'select doc_id, updated, size, error from documents where full_name=:1';
  QAddFile := IndexProvider.CreateQuery(true);
  QError := IndexProvider.CreateQuery(true);
  QError.SQL := 'update documents set error =:1 where doc_id=:2';
  QUpdateFile := IndexProvider.CreateQuery(true);
  QUpdateFile.SQL := 'update documents set updated=:1, size=:2, error=null where doc_id=:3';
  QU := IndexProvider.CreateQuery(true);
  QX := IndexProvider.CreateQuery;
  if IndexProvider.ProviderType = Firebird then
    QAddFile.SQL := 'insert into documents(full_name, location, file_type, updated, size, file_name, title, folder_id) values (:0, :1, :2, :3, :4, :5, :6, :7) returning doc_id'
  else
    QAddFile.SQL := 'insert into documents(full_name, location, file_type, updated, size, file_name, title, folder_id) values (:0, :1, :2, :3, :4, :5, :6, :7); select last_insert_rowid();';
end;

procedure TSQLiteDocumentDatabase.DeleteDocument(DocID: integer);
begin
  inherited;

end;

destructor TSQLiteDocumentDatabase.Destroy;
begin
  QFile.Free;
  QAddFile.Free;
  QError.Free;
  QUpdateFile.Free;
  QU.Free;
  QX.Free;
  FreeandNil(IndexProvider);
  FreeandNil(FProviderPool);
  FreeandNil(FAdapterPool);
  inherited;
end;

function TSQLiteDocumentDatabase.GetDocumentbyFileName(const AFileName: string;
  out DocID: integer; out AUpdated: TDateTime): boolean;
begin
  QFile.SelectWP([AFileName]);
  Result := not QFile.Eof;
  if Result then
  begin
    DocID := QFile.AsInteger(0);
    AUpdated := QFile.Dataset.Fields[1].AsDateTime;
  end;
end;

function TSQLiteDocumentDatabase.GetDocumentbyId(DocID: integer; var FullName,
  FileType, FileName, Title: string; var FolderId: integer): boolean;
var P: TDMProvider;
    Q: TDMQuery;
begin
  P := ProviderPool.GetProvider;
  try
    Q := P.CreateQuery;
    try
      Q.SelectWP([DocID], 'select full_name, file_type, file_name, title, folder_id from documents where doc_id=:1');
      Result := not  Q.Eof;
      if Result then
      begin
        FullName := Q.AsString(0);
        FileType := Q.AsString(1);
        FileName := Q.AsString(2);
        Title := Q.AsString(3);
        FolderId := Q.AsInteger(4);
      end;
    finally
      Q.Free
    end;
  finally
    ProviderPool.ReleaseProvider(P)
  end;
end;

procedure TSQLiteDocumentDatabase.GetFolders(out AFolders: THtDocumentFolders);
begin
  QX.Select('select path, folder_id, subdirs from folders order by folder_id');
  SetLength(AFolders, 0);
  while not QX.Eof do
  begin
    SetLength(AFolders, Length(AFolders) + 1);
    AFolders[High(AFolders)].ID := QX.AsInteger(1);
    AFolders[High(AFolders)].Path := QX.AsString(0);
    AFolders[High(AFolders)].Ignore := QX.AsInteger(2) = 0;
    QX.Next;
  end;
end;

procedure TSQLiteDocumentDatabase.SetDocumentError(DocID: integer;
  const AMessage: string);
begin
  QError.ExecWP([copy(AMessage, 1, 200), DocID]);
  QError.Commit;
end;

procedure TSQLiteDocumentDatabase.UpdateDocument(ADocument: THtStorageDoc);
begin
  QUpdateFile.ExecWP([ADocument.Created, ADocument.Size, ADocument.DocID]);
end;


end.
